﻿using System;
using System.Runtime.InteropServices;

namespace HeuristicLab.Problems.DynamicalSystemsModelling {
  public static class CVODES {

    // CVODES types
    public enum MultistepMethod : int { CV_ADAMS = 1, CV_BDF = 2 };
    public enum NonlinearSolverIteration : int { CV_NEWTON = 1, CV_FUNCTIONAL = 2 };


    /* itask */
    public const int CV_NORMAL = 1;
    public const int CV_ONE_STEP = 2;

    /* ism */
    public const int CV_SIMULTANEOUS = 1;
    public const int CV_STAGGERED = 2;
    public const int CV_STAGGERED1 = 3;

    /* DQtype */
    public const int CV_CENTERED = 1;
    public const int CV_FORWARD = 2;

    /* interp */
    public const int CV_HERMITE = 1;
    public const int CV_POLYNOMIAL = 2;

    /* 
     * ----------------------------------------
     * CVODES return flags
     * ----------------------------------------
     */

    public const int CV_SUCCESS = 0;
    public const int CV_TSTOP_RETURN = 1;
    public const int CV_ROOT_RETURN = 2;

    public const int CV_WARNING = 99;

    public const int CV_TOO_MUCH_WORK = -1;
    public const int CV_TOO_MUCH_ACC = -2;
    public const int CV_ERR_FAILURE = -3;
    public const int CV_CONV_FAILURE = -4;

    public const int CV_LINIT_FAIL = -5;
    public const int CV_LSETUP_FAIL = -6;
    public const int CV_LSOLVE_FAIL = -7;
    public const int CV_RHSFUNC_FAIL = -8;
    public const int CV_FIRST_RHSFUNC_ERR = -9;
    public const int CV_REPTD_RHSFUNC_ERR = -10;
    public const int CV_UNREC_RHSFUNC_ERR = -11;
    public const int CV_RTFUNC_FAIL = -12;
    public const int CV_CONSTR_FAIL = -13;

    public const int CV_MEM_FAIL = -20;
    public const int CV_MEM_NULL = -21;
    public const int CV_ILL_INPUT = -22;
    public const int CV_NO_MALLOC = -23;
    public const int CV_BAD_K = -24;
    public const int CV_BAD_T = -25;
    public const int CV_BAD_DKY = -26;
    public const int CV_TOO_CLOSE = -27;

    public const int CV_NO_QUAD = -30;
    public const int CV_QRHSFUNC_FAIL = -31;
    public const int CV_FIRST_QRHSFUNC_ERR = -32;
    public const int CV_REPTD_QRHSFUNC_ERR = -33;
    public const int CV_UNREC_QRHSFUNC_ERR = -34;

    public const int CV_NO_SENS = -40;
    public const int CV_SRHSFUNC_FAIL = -41;
    public const int CV_FIRST_SRHSFUNC_ERR = -42;
    public const int CV_REPTD_SRHSFUNC_ERR = -43;
    public const int CV_UNREC_SRHSFUNC_ERR = -44;

    public const int CV_BAD_IS = -45;

    public const int CV_NO_QUADSENS = -50;
    public const int CV_QSRHSFUNC_FAIL = -51;
    public const int CV_FIRST_QSRHSFUNC_ERR = -52;
    public const int CV_REPTD_QSRHSFUNC_ERR = -53;
    public const int CV_UNREC_QSRHSFUNC_ERR = -54;

    /* 
     * ----------------------------------------
     * CVODEA return flags
     * ----------------------------------------
     */

    public const int CV_NO_ADJ = -101;
    public const int CV_NO_FWD = -102;
    public const int CV_NO_BCK = -103;
    public const int CV_BAD_TB0 = -104;
    public const int CV_REIFWD_FAIL = -105;
    public const int CV_FWD_FAIL = -106;


    public const int CV_GETY_BADT = -107;

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVRhsFunc(
        double t, // realtype
        IntPtr y, // N_Vector
        IntPtr ydot, // N_Vector
        IntPtr user_data
      );

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVRhsFuncB(
        double t, // realtype
        IntPtr y, // N_Vector
        IntPtr yB, // N_Vector
        IntPtr yBdot, // N_Vector
        IntPtr user_data
      );

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVDlsJacFunc(
      double t,
      IntPtr y, // N_Vector
      IntPtr fy, // N_Vector
      IntPtr Jac, // SUNMatrix
      IntPtr user_data,
      IntPtr tmp1, // N_Vector
      IntPtr tmp2, // N_Vector
      IntPtr tmp3 // N_Vector
      );

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVDlsJacFuncB(
      double t,
      IntPtr y, // N_Vector
      IntPtr yB, // N_Vector
      IntPtr fyB, // N_Vector
      IntPtr Jac, // SUNMatrix
      IntPtr user_data,
      IntPtr tmp1, // N_Vector
      IntPtr tmp2, // N_Vector
      IntPtr tmp3 // N_Vector
    );

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVQuadRhsFn(
      double t,
      IntPtr y, // N_Vector
      IntPtr yQdot, // N_Vector
      IntPtr user_data
      );

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVQuadRhsFnB(
      double t,
      IntPtr y, // N_Vector
      IntPtr yB, // N_Vector
      IntPtr qBdot, // N_Vector
      IntPtr user_data
      );


    // to calculate sensitivities RHS for all equations at once
    // must compute (∂f/∂y)s_i(t) + ∂f/∂p_i
    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate int CVSensRhsFn(
      int Ns,
      double t,
      IntPtr y, // N_Vector
      IntPtr ydot, // N_Vector
      IntPtr yS, // N_Vector* one vector for each parameter
      IntPtr ySdot, // N_Vector* one vector for each parameter
      IntPtr user_data,
      IntPtr tmp1, // N_Vector
      IntPtr tmp2 // N_Vector
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeCreate", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    // returns a void* to the cvodes memory block if successful otherwise NULL
    public static extern IntPtr CVodeCreate(MultistepMethod lmm, NonlinearSolverIteration iter);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeInit", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeInit(
      IntPtr cvode_mem, // pointer returned by CVodeCreate
      CVRhsFunc f,
      double t0, // realtype, the inital value of t
      IntPtr y0 // N_Vector the initial value of y
    );


    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeSStolerances", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeSStolerances(
      IntPtr cvode_mem,
      double reltol,
      double abstol
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVDlsSetLinearSolver", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVDlsSetLinearSolver(
      IntPtr cvode_mem,
      IntPtr linearSolver, // SUNLinearSolver
      IntPtr j // SUNMatrix
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVDlsSetJacFn", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVDlsSetJacFn(
      IntPtr cvode_mem,
      CVDlsJacFunc jacFunc
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeSensInit", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeSensInit(
      IntPtr cvode_mem,
      int Ns, // number of parameters
      int ism, // sensitivity solution method, CV_SIMULTANEOUS or CV_STAGGERED
      CVSensRhsFn fS, // right hand side function which computes all sensitivity RHS at the same time
      IntPtr yS0 // N_Vector
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeSensSStolerances", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeSensSStolerances(
      IntPtr cvode_mem,
      double reltol,
      double[] abstol
    );

    /* Call CVodeSensEEtolerances to estimate tolerances for sensitivity 
   variables based on the rolerances supplied for states variables and 
   the scaling factor pbar */
    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeSensEEtolerances", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeSensEEtolerances(
      IntPtr cvode_mem
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetSens", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeGetSens(
      IntPtr cvode_mem,
      ref double tret,
      IntPtr yS //N_Vector*, one vector for each parameter
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeQuadInit", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeQuadInit(
      IntPtr cvode_mem,
      CVQuadRhsFn qF,
      IntPtr yQ0 // N_Vector, initial value of yQ
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeQuadInitB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeQuadInitB(
      IntPtr cvode_mem,
      int indexB,
      CVQuadRhsFnB rhsQB,
      IntPtr yQB0 // N_Vector, initial value of yQB
    );


    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeAdjInit", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeAdjInit(
      IntPtr cvode_mem,
      int nd,  // integration steps between checkpoints
      int interpType // either CV_POLYNOMIAL or CV_HERMITE
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeF", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeF(
        IntPtr cvode_mem,
        double t_out, // the next time at which a computed solution is desired
        IntPtr y, // N_Vector, the solution vector y
        ref double t_ret, // the time reached by the solver (output)
        int itask, // CV_NORMAL or CV_ONE_STEP
        ref int ncheck  // the number of internal checkpoints stored so far.
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetNumSteps", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeGetNumSteps(
      IntPtr cvode_mem,
      ref long ncheck  // the number of steps taken
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeCreateB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeCreateB(
        IntPtr cvode_mem,
        MultistepMethod lmmB,
        NonlinearSolverIteration iterB,
        ref int which
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeInitB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeInitB(
        IntPtr cvode_mem,
        int which,
        CVRhsFuncB rightHandSide,
        double tB0, // endpoint T where final conditions are provided (equal to endpoint of forward integration)
        IntPtr yB0 // N_Vector inital value at t=tb0 of backward solution
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVode", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVode(
      IntPtr cvode_mem,
      double tout, // next time at which a solution is desired
      IntPtr yout, // N_Vector
      ref double tret, // the time reached by the solver (output)
      int itask // flag indicating the job of the solver for the next step. 
      );


    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeSStolerancesB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeSStolerancesB(
      IntPtr cvode_mem,
      int which,
      double reltol,
      double abstol
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVDlsSetLinearSolverB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVDlsSetLinearSolverB(
      IntPtr cvode_mem,
      int which,
      IntPtr linearSolver, // SUNLinearSolver
      IntPtr j // SUNMatrix
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVDlsSetJacFnB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVDlsSetJacFnB(
      IntPtr cvode_mem,
      int which,
      CVDlsJacFuncB jacFunc
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeB(
      IntPtr cvode_mem,
      double tout, // next time at which a solution is desired
      int itask // flag indicating the job of the solver for the next step. 
    );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeGetB(
          IntPtr cvode_mem,
          int which,
          ref double tret,
          IntPtr yB // the backward solution at time tret
        );
    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetAdjY", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeGetAdjY(
          IntPtr cvode_mem,
          double t,
          IntPtr y // the forward solution y(t)
        );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetAdjCVodeBmem", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr CVodeGetAdjCVodeBmem(
              IntPtr cvode_mem,
              int which
            );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeGetQuadB", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int CVodeGetQuadB(
        IntPtr cvode_mem,
        int which,
        ref double tret,
        IntPtr yQB // N_Vector
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "CVodeFree", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]

    public static extern int CVodeFree(ref IntPtr cvode_mem);

    #region matrix
    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNDenseMatrix", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr SUNDenseMatrix(long m, long n);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNMatDestroy", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SUNMatDestroy(IntPtr A);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNDenseMatrix_Data", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public unsafe static extern double* SUNDenseMatrix_Data(IntPtr matrix);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNDenseMatrix_Cols", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern long SUNDenseMatrix_Cols(IntPtr matrix);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNDenseMatrix_Rows", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern long SUNDenseMatrix_Rows(IntPtr matrix);


    public unsafe static double SUNDenseMatrix_Get(IntPtr mat, long i, long j) {
      long M = SUNDenseMatrix_Rows(mat);
      // the(i, j)th element of A(with 0 <= i<M and 0 <= j<N) is given by(A->data)[j*M+i]
      return SUNDenseMatrix_Data(mat)[j * M + i];
    }

    public unsafe static void SUNDenseMatrix_Set(IntPtr mat, long i, long j, double val) {
      long M = SUNDenseMatrix_Rows(mat);
      // the(i, j)th element of A(with 0 <= i<M and 0 <= j<N) is given by(A->data)[j*M+i]
      SUNDenseMatrix_Data(mat)[j * M + i] = val;
    }

    #endregion

    #region linear solver
    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNDenseLinearSolver", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr SUNDenseLinearSolver(
      IntPtr y, // N_Vector
      IntPtr A // SUNMatrix
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNLinSolInitialize", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SUNLinSolInitialize(IntPtr linearSolver);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNLinSolSetup", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SUNLinSolSetup(
      IntPtr linearSolver,
      IntPtr A // SUNMatrix
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNLinSolSolve", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SUNLinSolSolve(
      IntPtr linearSolver,
      IntPtr A, // SUNMatrix
      IntPtr x, // N_Vector
      IntPtr b, // N_Vector
      double tol
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "SUNLinSolFree", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SUNLinSolFree(IntPtr linearSolver);

    #endregion

    #region N_Vector
    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VNew_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr N_VNew_Serial(long vec_length);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VDestroy_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern void N_VDestroy_Serial(IntPtr vec);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VDestroyVectorArray_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern void N_VDestroyVectorArray_Serial(IntPtr vecArr, int count);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VPrint_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern void N_VPrint_Serial(IntPtr vec);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VConst_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern void N_VConst_Serial(double c, IntPtr vec);


    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VL1Norm_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern double N_VL1Norm_Serial(IntPtr vec);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VMake_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr N_VMake_Serial(long vec_length, double[] v_data);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VScale", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    ///  Performs the operation z = c*x
    public static extern void N_VScale(double s,
      IntPtr x, // N_Vector
      IntPtr z // N_Vector
      );

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VMake_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public unsafe static extern double* N_VGetArrayPointer_Serial(IntPtr vec);

    [DllImport("sundials_cvodes-x86.dll", EntryPoint = "N_VCloneVectorArray_Serial", ExactSpelling = true, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr N_VCloneVectorArray_Serial(int count, IntPtr vec); // returns N_Vector* !


    /*
#define NV_CONTENT_S(v)  ( (N_VectorContent_Serial)(v->content) )

#define NV_LENGTH_S(v)   ( NV_CONTENT_S(v)->length )

#define NV_OWN_DATA_S(v) ( NV_CONTENT_S(v)->own_data )

#define NV_DATA_S(v)     ( NV_CONTENT_S(v)->data )

#define NV_Ith_S(v,i)    ( NV_DATA_S(v)[i] )
*/
    // methods for macros
    public unsafe static int* NV_CONTENT_S(IntPtr v) {
      int* content = (int*)*(int*)v.ToPointer();
      return content;
    }

    public unsafe static long NV_LENGTH_S(IntPtr v) {
      long length = *NV_CONTENT_S(v);
      return length;
    }
    public unsafe static bool NV_OWN_DATA_S(IntPtr v) {
      var content = NV_CONTENT_S(v);
      int own_data = *(content + 2);
      return own_data > 0;
    }
    public unsafe static double* NV_DATA_S(IntPtr v) {
      var content = NV_CONTENT_S(v);
      double* data = (double*)*(content + 3);
      return data;
    }
    public unsafe static double NV_Get_Ith_S(IntPtr v, long i) {
      return NV_DATA_S(v)[i];
    }
    public unsafe static void NV_Set_Ith_S(IntPtr v, long i, double val) {
      NV_DATA_S(v)[i] = val;
    }
    #endregion


  }
}
